Namespace NaiveBayes
	Friend Class NaiveBayesProgram
		Private Shared ran As New Random(25) ' 25 gives a nice demo

		Shared Sub Main(ByVal args() As String)
			Try
				Console.WriteLine(vbCrLf & "Begin Naive Bayes Classification demo" & vbCrLf)
				Console.WriteLine("Demo will classify sex based on occpation, dominance, height" & vbCrLf)

				Dim attributes() As String = { "occupation", "dominance", "height", "sex" }

				Dim attributeValues(attributes.Length - 1)() As String ' could scan values from raw data
				attributeValues(0) = New String() { "administrative", "construction", "education", "technology" }
				attributeValues(1) = New String() { "left", "right" }
				attributeValues(2) = New String() { "short", "medium", "tall" }
				attributeValues(3) = New String() { "male", "female" }

				Dim numericAttributeBorders(0) As Double ' there may be several numeric variables
				numericAttributeBorders(0) = New Double() { 64.0, 71.0 } ' height range: 57.0 to 78.0

				Console.WriteLine("Generating 40 lines of occupation, dominance, height, sex data" & vbCrLf)
				Dim data() As String = MakeData(40)

				Console.WriteLine("First 4 lines of training data are:" & vbCrLf)

				For i  = 0 To 3
					Console.WriteLine(data(i))
				Next i
				Console.WriteLine(vbCrLf)

				Console.WriteLine("Converting numeric height data to categorical data on 64.0 71.0" & vbCrLf)

				Dim binnedData() As String = BinData(data, attributeValues, numericAttributeBorders) ' convert numeric heights to categories

				Console.WriteLine("First 4 lines of binned training data are:" & vbCrLf)
				For i  = 0 To 3
					Console.WriteLine(binnedData(i))
				Next i
				Console.WriteLine(vbCrLf)

				Console.WriteLine("Scanning binned data to compute joint and dependent counts" & vbCrLf)
				Dim jointCounts As Integer = MakeJointCounts(binnedData, attributes, attributeValues)
				Dim dependentCounts() As Integer = MakeDependentCounts(jointCounts, 2)
				Console.WriteLine("Total male = " & dependentCounts(0))
				Console.WriteLine("Total female = " & dependentCounts(1))
				Console.WriteLine("")

				ShowJointCounts(jointCounts, attributeValues)

				' classify the sex of a person whose occupation is education, is right-handed, and is tall
				Dim occupation As String = "education"
				Dim dominance As String = "right"
				Dim height As String = "tall"

				Dim withLaplacian As Boolean = True ' prevent joint counts with 0

				Console.WriteLine("Using Naive Bayes " & (If(withLaplacian, "with", "without")) & " Laplacian smoothing to classify when:")
				Console.WriteLine(" occupation = " & occupation)
				Console.WriteLine(" dominance = " & dominance)
				Console.WriteLine(" height = " & height)
				Console.WriteLine("")

				Dim c As Integer = Classify(occupation, dominance, height, jointCounts, dependentCounts, withLaplacian, 3)
				If c = 0 Then
					Console.WriteLine(vbCrLf & "Data case is most likely male")
				ElseIf c = 1 Then
					Console.WriteLine(vbCrLf & "Data case is most likely female")
				End If

				Console.WriteLine(vbCrLf & "End demo" & vbCrLf)
				Console.ReadLine()
			Catch ex As Exception
				Console.WriteLine(ex.Message)
				Console.ReadLine()
			End Try
		End Sub ' Main

		Private Shared Function MakeData(ByVal numRows As Integer) As String() ' make dummy data
			Dim result(numRows - 1) As String
			For i  = 0 To numRows - 1
				Dim sex As String = MakeSex()
				Dim occupation As String = MakeOccupation(sex)
				Dim dominance As String = MakeDominance(sex)
				Dim height As String = MakeHeight(sex)
				Dim s As String = occupation & "," & dominance & "," & height & "," & sex
				result(i) = s
			Next i
			Return result
		End Function

		Private Shared Function MakeSex() As String
			Dim r As Integer = ran.Next(0, 19)
			If r >= 0 AndAlso r <= 11 Then ' 60%
				Return "male"
			ElseIf r >= 12 AndAlso r <= 19 Then ' 40%
				Return "female"
			End If
			Return "error"
		End Function

		Private Shared Function MakeDominance(ByVal sex As String) As String
			Dim p As Double = ran.NextDouble
			If sex = "male" Then
				If p < 0.33 Then
					Return "left"
					Else
						Return "right"
					End If
			ElseIf sex = "female" Then
				If p < 0.20 Then
					Return "left"
					Else
						Return "right"
					End If
			End If
			Return "error"
		End Function

		Private Shared Function MakeOccupation(ByVal sex As String) As String
			Dim r As Integer = ran.Next(0, 20)
			If sex = "male" Then
				If r = 0 Then ' 5%
					Return "administrative"
				ElseIf r >= 1 AndAlso r <= 6 Then ' 30%
					Return "construction"
				ElseIf r >= 7 AndAlso r <= 9 Then ' 15%
					Return "education"
				ElseIf r >= 10 AndAlso r <= 19 Then ' 50%
					Return "technology"
				End If
			ElseIf sex = "female" Then
                If r >= 0 AndAlso r <= 9 Then ' 50%
                    Return "administrative"
                ElseIf r = 10 Then ' 5%
                    Return "construction"
                ElseIf r >= 11 AndAlso r <= 15 Then ' 25%
                    Return "education"
                ElseIf r >= 16 AndAlso r <= 19 Then ' 20%
                    Return "technology"
                End If
			End If
			Return "error"
		End Function

		Private Shared Function MakeHeight(ByVal sex As String) As String
			Dim bucket As Integer = 0 ' height bucket: 0 = short, 1 = medium, 2 = tall
			Dim p As Double = ran.NextDouble
			If p < 0.1587 Then
				bucket = 0
			ElseIf p > 0.8413 Then
				bucket = 2
			Else ' p = (2 * 0.3413) = 0.6826
				bucket = 1
			End If

			Dim hi As Double = 0.0
			Dim lo As Double = 0.0

			If sex = "male" Then
				If bucket = 0 Then
					lo = 60.0
					hi = 66.0
				ElseIf bucket = 1 Then
					lo = 66.0
					hi = 72.0
				ElseIf bucket = 2 Then
					lo = 72.0
					hi = 78.0
				End If
			ElseIf sex = "female" Then
				If bucket = 0 Then
					lo = 57.0
					hi = 63.0
				ElseIf bucket = 1 Then
					lo = 63.0
					hi = 69.0
				ElseIf bucket = 2 Then
					lo = 69.0
					hi = 75.0
				End If
			End If

			Dim resultAsDouble As Double = (hi - lo) * ran.NextDouble + lo
			Return resultAsDouble.ToString("F1")
		End Function

		Private Shared Function BinData(ByVal data() As String, ByVal attributeValues As String, ByVal numericAttributeBorders As Double) As String()
			' convert numeric height to "short", "medium", or "tall". assumes data is occupation,dominance,height,sex
			Dim result(data.Length - 1) As String
			Dim tokens() As String
			Dim heightAsDouble As Double
			Dim heightAsBinnedString As String

			For i  = 0 To data.Length - 1
				tokens = data(i).Split(","c)
				heightAsDouble = Double.Parse(tokens(2))
				If heightAsDouble <= numericAttributeBorders(0)(0) Then ' short
					heightAsBinnedString = attributeValues(2)(0)
				ElseIf heightAsDouble >= numericAttributeBorders(0)(1) Then ' tall
					heightAsBinnedString = attributeValues(2)(2)
				Else
					heightAsBinnedString = attributeValues(2)(1) ' medium
				End If

				Dim s As String = tokens(0) & "," & tokens(1) & "," & heightAsBinnedString & "," & tokens(3)
				result(i) = s
			Next i
			Return result
		End Function

		Private Shared Function MakeJointCounts(ByVal binnedData() As String, ByVal attributes() As String, ByVal attributeValues As String) As Integer
			' assumes binned data is occupation, dominance, height, sex
			' result][] -> [attributeatt valuesex
			' ex: result031 is the count of (occupation) (technology) (female), i.e., the count of technology AND female

			Dim jointCounts(attributes.Length - 2) As Integer ' note the -1 (no sex)

			jointCounts(0) = New Integer(3){} ' 4 occupations
			jointCounts(1) = New Integer(1){} ' 2 dominances
			jointCounts(2) = New Integer(2){} ' 3 heights

			jointCounts(0)(0) = New Integer(1){} ' 2 sexes for administrative
			jointCounts(0)(1) = New Integer(1){} ' construction
			jointCounts(0)(2) = New Integer(1){} ' education
			jointCounts(0)(3) = New Integer(1){} ' tedchnology

			jointCounts(1)(0) = New Integer(1){} ' left
			jointCounts(1)(1) = New Integer(1){} ' right

			jointCounts(2)(0) = New Integer(1){} ' short
			jointCounts(2)(1) = New Integer(1){} ' medium
			jointCounts(2)(2) = New Integer(1){} ' tall

			For i  = 0 To binnedData.Length - 1
				Dim tokens() As String = binnedData(i).Split(","c)

				Dim occupationIndex As Integer = AttributeValueToIndex(0, tokens(0))
				Dim dominanceIndex As Integer = AttributeValueToIndex(1, tokens(1))
				Dim heightIndex As Integer = AttributeValueToIndex(2, tokens(2))
				Dim sexIndex As Integer = AttributeValueToIndex(3, tokens(3))

                jointCounts(0)(occupationIndex)(sexIndex) = jointCounts(0)(occupationIndex)(sexIndex) + 1 ' occupation and sex count
                jointCounts(1)(dominanceIndex)(sexIndex) = jointCounts(1)(dominanceIndex)(sexIndex) + 1
                jointCounts(2)(heightIndex)(sexIndex) = jointCounts(2)(heightIndex)(sexIndex) + 1
			Next i

			Return jointCounts
		End Function

		Private Shared Function AttributeValueToIndex(ByVal attribute As Integer, ByVal attributeValue As String) As Integer
			' we could do this programmatically (maybe with a Dictionary) but a crude approach is more clear
			If attribute = 0 Then
				If attributeValue = "administrative" Then
					Return 0
				ElseIf attributeValue = "construction" Then
					Return 1
				ElseIf attributeValue = "education" Then
					Return 2
				ElseIf attributeValue = "technology" Then
					Return 3
				End If
			ElseIf attribute = 1 Then
				If attributeValue = "left" Then
					Return 0
				ElseIf attributeValue = "right" Then
					Return 1
				End If
			ElseIf attribute = 2 Then
				If attributeValue = "short" Then
					Return 0
				ElseIf attributeValue = "medium" Then
					Return 1
				ElseIf attributeValue = "tall" Then
					Return 2
				End If
			ElseIf attribute = 3 Then
				If attributeValue = "male" Then
					Return 0
				ElseIf attributeValue = "female" Then
					Return 1
				End If
			End If
			Return -1 ' error
		End Function

		Private Shared Sub ShowJointCounts(ByVal jointCounts As Integer, ByVal attributeValues As String)
			For k  = 0 To 1
				For i  = 0 To jointCounts.Length - 1
					For j  = 0 To jointCounts(i).Length - 1
						Console.WriteLine(attributeValues(i)(j).PadRight(15) & "& " & attributeValues(3)(k).PadRight(6) & " = " & jointCounts(i)(j)(k))
					Next j
				Next i
				Console.WriteLine("") ' separate sexes
			Next k
		End Sub

		Private Shared Function MakeDependentCounts(ByVal jointCounts As Integer, ByVal numDependents As Integer) As Integer()
			Dim result(numDependents - 1) As Integer
			For k  = 0 To numDependents - 1 ' male then female
				For j  = 0 To jointCounts(0).Length - 1 ' scanning attribute 0 = occupation. could use any attribute
                    result(k) = result(k) + jointCounts(0)(j)(k)
				Next j
			Next k

			Return result
		End Function

		Private Shared Function Classify(ByVal occupation As String, ByVal dominance As String, ByVal height As String, ByVal jointCounts As Integer, ByVal dependentCounts() As Integer, ByVal withSmoothing As Boolean, ByVal xClasses As Integer) As Integer
			Dim partProbMale As Double = PartialProbability("male", occupation, dominance, height, jointCounts, dependentCounts, withSmoothing, xClasses)
			Dim partProbFemale As Double = PartialProbability("female", occupation, dominance, height, jointCounts, dependentCounts, withSmoothing, xClasses)
			Dim evidence As Double = partProbMale + partProbFemale
			Dim probMale As Double = partProbMale / evidence
			Dim probFemale As Double = partProbFemale / evidence

			'Console.WriteLine("Partial prob of male   = " + partProbMale.ToString("F6"));
			'Console.WriteLine("Partial prob of female = " + partProbFemale.ToString("F6"));

			Console.WriteLine("Probability of male   = " & probMale.ToString("F4"))
			Console.WriteLine("Probability of female = " & probFemale.ToString("F4"))
			If probMale > probFemale Then ' could use a threshold
				Return 0
			Else
				Return 1
			End If
		End Function

		Private Shared Function PartialProbability(ByVal sex As String, ByVal occupation As String, ByVal dominance As String, ByVal height As String, ByVal jointCounts As Integer, ByVal dependentCounts() As Integer, ByVal withSmoothing As Boolean, ByVal xClasses As Integer) As Double
			Dim sexIndex As Integer = AttributeValueToIndex(3, sex)

			Dim occupationIndex As Integer = AttributeValueToIndex(0, occupation)
			Dim dominanceIndex As Integer = AttributeValueToIndex(1, dominance)
			Dim heightIndex As Integer = AttributeValueToIndex(2, height)

			Dim totalMale As Integer = dependentCounts(0)
			Dim totalFemale As Integer = dependentCounts(1)
			Dim totalCases As Integer = totalMale + totalFemale

			Dim totalToUse As Integer = 0
			If sex = "male" Then
				totalToUse = totalMale
			ElseIf sex = "female" Then
				totalToUse = totalFemale
			End If

			Dim p0 As Double = (totalToUse * 1.0) / (totalCases) ' prob of either male or female
			Dim p1 As Double = 0.0
			Dim p2 As Double = 0.0
			Dim p3 As Double = 0.0

			If withSmoothing = False Then
				p1 = (jointCounts(0)(occupationIndex)(sexIndex) * 1.0) / totalToUse ' conditional prob of male (or female, depending on sex parameter) given the occupation
				p2 = (jointCounts(1)(dominanceIndex)(sexIndex) * 1.0) / totalToUse ' conditional prob of the specified sex, given the specified domnance
				p3 = (jointCounts(2)(heightIndex)(sexIndex) * 1.0) / totalToUse ' condition prob given specified height
			ElseIf withSmoothing = True Then ' Laplacian smoothing to avoid 0-count joint probabilities
				p1 = (jointCounts(0)(occupationIndex)(sexIndex) + 1) / ((totalToUse + xClasses) * 1.0) ' add 1 to count in numerator, add number x classes in denominator
				p2 = (jointCounts(1)(dominanceIndex)(sexIndex) + 1) / ((totalToUse + xClasses) * 1.0) ' conditional prob of the specified sex, given the specified domnance
				p3 = (jointCounts(2)(heightIndex)(sexIndex) + 1) / ((totalToUse + xClasses) * 1.0)
			End If

			'return p0 * p1 * p2 * p3; // risky if any very small values
			Return Math.Exp(Math.Log(p0) + Math.Log(p1) + Math.Log(p2) + Math.Log(p3))
		End Function

		Private Shared Function AnalyzeJointCounts(ByVal jointCounts As Integer) As Integer
			' check for any joint-counts that are 0 which could blow up Naive Bayes
			Dim zeroCounts As Integer = 0

			For i  = 0 To jointCounts.Length - 1 ' attribute
				For j  = 0 To jointCounts(i).Length - 1 ' value
					For k  = 0 To jointCounts(i)(j).Length - 1 ' sex
						If jointCounts(i)(j)(k) = 0 Then
                            zeroCounts = zeroCounts + 1
						End If
					Next k
				Next j
			Next i
			Return zeroCounts
		End Function
	End Class ' class Program
End Namespace ' ns
